/*
    FreeRTOS V6.0.1 - Copyright (C) 2009 Real Time Engineers Ltd.

    ***************************************************************************
    *                                                                         *
    * If you are:                                                             *
    *                                                                         *
    *    + New to FreeRTOS,                                                   *
    *    + Wanting to learn FreeRTOS or multitasking in general quickly       *
    *    + Looking for basic training,                                        *
    *    + Wanting to improve your FreeRTOS skills and productivity           *
    *                                                                         *
    * then take a look at the FreeRTOS eBook                                  *
    *                                                                         *
    *        "Using the FreeRTOS Real Time Kernel - a Practical Guide"        *
    *                  http://www.FreeRTOS.org/Documentation                  *
    *                                                                         *
    * A pdf reference manual is also available.  Both are usually delivered   *
    * to your inbox within 20 minutes to two hours when purchased between 8am *
    * and 8pm GMT (although please allow up to 24 hours in case of            *
    * exceptional circumstances).  Thank you for your support!                *
    *                                                                         *
    ***************************************************************************

    This file is part of the FreeRTOS distribution.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation AND MODIFIED BY the FreeRTOS exception.
    ***NOTE*** The exception to the GPL is included to allow you to distribute
    a combined work that includes FreeRTOS without being obliged to provide the
    source code for proprietary components outside of the FreeRTOS kernel.
    FreeRTOS is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details. You should have received a copy of the GNU General Public 
    License and the FreeRTOS license exception along with FreeRTOS; if not it 
    can be viewed here: http://www.freertos.org/a00114.html and also obtained 
    by writing to Richard Barry, contact details for whom are available on the
    FreeRTOS WEB site.

    1 tab == 4 spaces!

    http://www.FreeRTOS.org - Documentation, latest information, license and
    contact details.

    http://www.SafeRTOS.com - A version that is certified for use in safety
    critical systems.

    http://www.OpenRTOS.com - Commercial support, development, porting,
    licensing and training services.
*/

/*
 * Interrupt driven driver for the EMAC peripheral.  This driver is not
 * reentrant, re-entrancy is handled by a semaphore at the network interface
 * level. 
 */


/*
Changes from V3.2.2

	+ Corrected the byte order when writing the MAC address to the MAC.
	+ Support added for MII interfaces.  Previously only RMII was supported.

Changes from V3.2.3

	+ The MII interface is now the default.
	+ Modified the initialisation sequence slightly to allow auto init more
	  time to complete.

Changes from V4.0.1

	+ Made the function vClearEMACTxBuffer() more robust by moving the index
	  manipulation into the if() statement.  This allows the tx interrupt to
	  execute even when there is no data to handle.

Changes from V4.0.4

	+ Corrected the Rx frame length mask when obtaining the length from the
	  rx descriptor.
*/


/* Standard includes. */
#include <string.h>

/* Scheduler includes. */
#include "FreeRTOS.h"
#include "semphr.h"
#include "task.h"

/* Demo app includes. */
#include "lpc23xx.h"
#include "LPC24xx_EMAC.h"

/* Hardware specific includes. */
#include "Emac.h"
//#include "mii.h"


/* A short delay is used to wait for a buffer to become available, should
one not be immediately available when trying to transmit a frame. */
#define emacBUFFER_WAIT_DELAY		( 2 )
#define emacMAX_WAIT_CYCLES			( ( portBASE_TYPE ) ( configTICK_RATE_HZ / 40 ) )

/* The time to block waiting for input. */
#define emacBLOCK_TIME_WAITING_FOR_INPUT	( ( portTickType ) 100 )

/* Misc defines. */
#define emacINTERRUPT_LEVEL			( 5 )
#define emacNO_DELAY				( 0 )
#define emacTOTAL_FRAME_HEADER_SIZE	( 54 )
#define emacPHY_INIT_DELAY			( 5000 / portTICK_RATE_MS )
#define emacRESET_KEY				( ( unsigned long ) 0xA5000000 )
#define emacRESET_LENGTH			( ( unsigned long ) ( 0x01 << 8 ) )

/*-----------------------------------------------------------*/

/* The IP and Ethernet addresses are read from the header files. */
const char cMACAddress[ 6 ] = { emacETHADDR0, emacETHADDR1, emacETHADDR2, emacETHADDR3, emacETHADDR4, emacETHADDR5 };
const unsigned char ucIPAddress[ 4 ]  = { emacIPADDR0, emacIPADDR1, emacIPADDR2, emacIPADDR3 };

/*-----------------------------------------------------------*/

/* See the header file for descriptions of public functions. */

/*
 * Prototype for the EMAC interrupt function.
 */
void vEMACISR_Wrapper( void ) __attribute__ ((naked));


/*
 * Configure the EMAC and AIC for EMAC interrupts.
 */
static void prvSetupEMACInterrupt( void );

void tx_descr_init (void);
void rx_descr_init (void);
void write_PHY (int PhyReg, int Value);
unsigned short read_PHY (unsigned char PhyReg);

/* The semaphore used by the EMAC ISR to wake the EMAC task. */
static xSemaphoreHandle xSemaphore = NULL;

/* Holds the index to the next buffer from which data will be read. */
static volatile unsigned long ulNextRxBuffer = 0;

/*-----------------------------------------------------------*/

/* See the header file for descriptions of public functions. */
long lEMACSend( char *pcFrom, unsigned long ulLength, long lEndOfFrame ){
	long lReturn = pdPASS;
	unsigned int  	Index;
	portBASE_TYPE	xWaitCycles = 0;
	unsigned int	IndexNext = MAC_TXPRODUCEINDEX + 1;
	static int 		offset = 0;

	if(ulLength == 0){
		return lReturn;
	}
	if(IndexNext > MAC_TXDESCRIPTORNUM){
		IndexNext = 0;
	}

	while(IndexNext == MAC_TXCONSUMEINDEX){
		if( ++xWaitCycles > emacMAX_WAIT_CYCLES ){
			/* Give up. */
			// lReturn = pdFAIL;
			break;
		}else{
			vTaskDelay( emacBUFFER_WAIT_DELAY );
		}
	}
	Index = MAC_TXPRODUCEINDEX;
	if (ulLength > ETH_FRAG_SIZE){
		ulLength = ETH_FRAG_SIZE;
	}
	/* lReturn will only be pdPASS if a buffer is available. */
	if( lReturn == pdPASS )
	{
		portENTER_CRITICAL();
		{
			memcpy((unsigned int *)(TX_BUF(Index) + offset), pcFrom, ulLength);
			offset += ulLength;

			if( lEndOfFrame ){
				TX_DESC_CTRL(Index) = ((offset - 1) & 0x7ff) | TCTRL_LAST;
				TX_STAT_INFO(Index) = 0;
				MAC_TXPRODUCEINDEX = IndexNext;
				offset = 0;
			}
		}
		portEXIT_CRITICAL();
	}

	return lReturn;
}
/*-----------------------------------------------------------*/

/* See the header file for descriptions of public functions. */
unsigned long ulEMACInputLength( void )
{

	if( ulNextRxBuffer != MAC_RXPRODUCEINDEX )
	{
		return (RX_STAT_INFO(ulNextRxBuffer) & 0x7ff)+1;
	}

	return 0;
}
/*-----------------------------------------------------------*/

/* See the header file for descriptions of public functions. */
void vEMACRead( char *pcTo, unsigned long ulSectionLength, unsigned long ulTotalFrameLength )
{
static unsigned long ulSectionBytesReadSoFar = 0, ulBufferPosition = 0, ulFameBytesReadSoFar = 0;
static char *pcSource;
register unsigned long ulBytesRemainingInBuffer, ulRemainingSectionBytes;

	/* Read ulSectionLength bytes from the Rx buffers.  This is not necessarily any
	correspondence between the length of our Rx buffers, and the length of the
	data we are returning or the length of the data being requested.  Therefore, 
	between calls  we have to remember not only which buffer we are currently 
	processing, but our position within that buffer.  This would be greatly
	simplified if PBUF_POOL_BUFSIZE could be guaranteed to be greater than
	the size of each Rx buffer, and that memory fragmentation did not occur.
	
	This function should only be called after a call to ulEMACInputLength().
	This will ensure ulNextRxBuffer is set to the correct buffer. */



	/* vEMACRead is called with pcTo set to NULL to indicate that we are about
	to read a new frame.  Any fragments remaining in the frame we were 
	processing during the last call should be dropped. */
	if( pcTo == NULL )
	{
		/* How many bytes are indicated as being in this buffer?  If none then
		the buffer is completely full and the frame is contained within more
		than one buffer. */

		/* Reset our state variables ready for the next read from this buffer. */
        pcSource = (char *)RX_BUF(ulNextRxBuffer);
        ulFameBytesReadSoFar = ( unsigned long ) 0;
		ulBufferPosition = ( unsigned long ) 0;
	}
	else
	{
		/* Loop until we have obtained the required amount of data. */
        ulSectionBytesReadSoFar = 0;
		while( ulSectionBytesReadSoFar < ulSectionLength )
		{
			/* We may have already read some data from this buffer.  How much
			data remains in the buffer? */
			ulBytesRemainingInBuffer = ( ETH_FRAG_SIZE - ulBufferPosition );

			/* How many more bytes do we need to read before we have the 
			required amount of data? */
            ulRemainingSectionBytes = ulSectionLength - ulSectionBytesReadSoFar;

			/* Do we want more data than remains in the buffer? */
			if( ulRemainingSectionBytes > ulBytesRemainingInBuffer )
			{
				/* We want more data than remains in the buffer so we can 
				write the remains of the buffer to the destination, then move
				onto the next buffer to get the rest. */
				memcpy( &( pcTo[ ulSectionBytesReadSoFar ] ), &( pcSource[ ulBufferPosition ] ), ulBytesRemainingInBuffer );
				ulSectionBytesReadSoFar += ulBytesRemainingInBuffer;
                ulFameBytesReadSoFar += ulBytesRemainingInBuffer;

				/* Move onto the next buffer. */
				ulNextRxBuffer++;
				if( ulNextRxBuffer > MAC_RXDESCRIPTORNUM )
				{
					ulNextRxBuffer = ( unsigned long ) 0;
				}
				MAC_RXCONSUMEINDEX = ulNextRxBuffer;

				/* Reset the variables for the new buffer. */
		        pcSource = (char *)RX_BUF(ulNextRxBuffer);
				ulBufferPosition = ( unsigned long ) 0;
			}
			else
			{
				/* We have enough data in this buffer to send back.  Read out
				enough data and remember how far we read up to. */
				memcpy( &( pcTo[ ulSectionBytesReadSoFar ] ), &( pcSource[ ulBufferPosition ] ), ulRemainingSectionBytes );

				/* There may be more data in this buffer yet.  Increment our 
				position in this buffer past the data we have just read. */
				ulBufferPosition += ulRemainingSectionBytes;
				ulSectionBytesReadSoFar += ulRemainingSectionBytes;
                ulFameBytesReadSoFar += ulRemainingSectionBytes;

				/* Have we now finished with this buffer? */
				if( ( ulBufferPosition >= ETH_FRAG_SIZE ) || ( ulFameBytesReadSoFar >= ulTotalFrameLength ) )
				{
					/* Move onto the next buffer. */
					ulNextRxBuffer++;
					if( ulNextRxBuffer > MAC_RXDESCRIPTORNUM )
					{
						ulNextRxBuffer = ( unsigned long ) 0;
					}
					MAC_RXCONSUMEINDEX = ulNextRxBuffer;

	
			        pcSource = (char *)RX_BUF(ulNextRxBuffer);
					ulBufferPosition = 0;
				}
			}
		}
	}
}
/*-----------------------------------------------------------*/

/* See the header file for descriptions of public functions. */
xSemaphoreHandle xEMACInit( void )
{
	// Initialize the EMAC ethernet controller.
	portBASE_TYPE xReturn = pdPASS;
	volatile unsigned int regv,tout,id1,id2;

	// Power Up the EMAC controller.
	PCONP |= 0x40000000;

	// Enable P1 Ethernet Pins.
	// p1.0(ENET_TXD0), 1(ENET_TXD1), 4(ENET_TX_EN), 6(ENET_TX_CLK), 8(ENET_CRS_DV/ENET_CRS),
	//    9(ENET_RXD0), 10(ENET_RXD1), 14(ENET_RX_ER), 15(ENET_REF_CLK/ENET_RX_CLK)
	// func=01, mode=10 (neither pull-up nor pull-down)
	PINSEL2 &= ~0xF03F330F;
	PINMODE2&= ~0xF03F330F;

	if (MAC_MODULEID == OLD_EMAC_MODULE_ID) {
		// For the first silicon rev.'-' ID P1.6 should be set.
		PINSEL2 |= 0x50151105;
		PINMODE2|= 0xA02A220A;
	}else {
		// on rev. 'A' and later, P1.6 should NOT be set.
		PINSEL2 |= 0x50150105;
		PINMODE2|= 0xA02A020A;
	}
	// p1.16(ENET_MDC), 17(ENET_MDIO), func=01, mode=10 (neither pull-up nor pull-down)
	PINSEL3 &= ~0x0000000F;
	PINMODE2&= ~0x0000000F;
	PINSEL3 |=  0x00000005;
	PINMODE2|=  0x0000000A;

	// Reset all EMAC internal modules.
	MAC_MAC1 = MAC1_RES_TX | MAC1_RES_MCS_TX | MAC1_RES_RX | MAC1_RES_MCS_RX | MAC1_SIM_RES | MAC1_SOFT_RES;
	MAC_COMMAND = CR_REG_RES | CR_TX_RES | CR_RX_RES;

	// A short delay after reset.
	for (tout = 100; tout; tout--);

	// Initialize MAC control registers.
	MAC_MAC1 = MAC1_PASS_ALL;
	MAC_MAC2 = MAC2_CRC_EN | MAC2_PAD_EN;
	MAC_MAXF = ETH_MAX_FLEN;
	MAC_CLRT = CLRT_DEF;
	MAC_IPGR = IPGR_DEF;

	//PCLK=18MHz, clock select=6, MDC=18/6=3MHz
	MAC_MCFG = MCFG_CLK_DIV20 | MCFG_RES_MII;
	for (tout = 100; tout; tout--);
	MAC_MCFG = MCFG_CLK_DIV20;

	// Enable Reduced MII interface.
	MAC_COMMAND = CR_RMII | CR_PASS_RUNT_FRM | CR_PASS_RX_FILT;

	// Reset Reduced MII Logic.
	MAC_SUPP = SUPP_RES_RMII | SUPP_SPEED;
	for (tout = 100; tout; tout--);
	MAC_SUPP = SUPP_SPEED;

	// Put the PHY in reset mode
	write_PHY (PHY_REG_BMCR, 0x8000);
	for (tout = 1000; tout; tout--);

	// Wait for hardware reset to end.
	for (tout = 0; tout < 0x100000; tout++) {
		regv = read_PHY (PHY_REG_BMCR);
		// Check 'RESET' bit and 'Power Down' bit
		if (!(regv & 0x8800)){
			// Reset complete
			break;
		}
	}

	// Check if this is a KSZ8721 PHY.
	id1 = read_PHY (PHY_REG_IDR1);
	id2 = read_PHY (PHY_REG_IDR2);

	if (((id1 << 16) | (id2 & 0xFFF0)) != KSZ8721_ID){
		xReturn = pdFAIL;
	}

	// Configure the PHY device
#if defined (_10MBIT_)
	// Connect at 10MBit
	write_PHY (PHY_REG_BMCR, PHY_FULLD_10M);
#elif defined (_100MBIT_)
	// Connect at 100MBit
	write_PHY (PHY_REG_BMCR, PHY_FULLD_100M);
#else
	// Use autonegotiation about the link speed.
	write_PHY (PHY_REG_BMCR, PHY_AUTO_NEG);
	// Wait to complete Auto_Negotiation.
	for (tout = 0; tout < 0x100000; tout++) {
		regv = read_PHY (PHY_REG_BMSR);
		if (regv & 0x0020) {
			// Autonegotiation Complete.
            break;
		}
	}
#endif
	if (tout >= 0x100000){
		xReturn = pdFAIL; // auto_neg failed
	}

	// Check the link status.
	for (tout = 0; tout < 0x10000; tout++) {
		// check if it is still in 'Still in auto-negotiation' state
		regv = read_PHY (PHY_REG_100TPCR);
		if (regv & 0x001C){
			// Link is on.
			xReturn = pdPASS;
			break;
		}
	}
	if (tout >= 0x10000){
		xReturn = pdFAIL;
	}

	if(xReturn == pdPASS){
		// Configure Full/Half Duplex mode.
		if (regv & 0x0010){
			// Full duplex is enabled.
			MAC_MAC2    |= MAC2_FULL_DUP;
			MAC_COMMAND |= CR_FULL_DUP;
			MAC_IPGT     = IPGT_FULL_DUP;
		}else {
			// Half duplex mode.
			MAC_IPGT = IPGT_HALF_DUP;
		}

		// Configure 100MBit/10MBit mode.
		if (regv & 0x0004) {
			// 10MBit mode.
			MAC_SUPP = 0;
		}else {
			// 100MBit mode.
			MAC_SUPP = SUPP_SPEED;
		}

		// Set the Ethernet MAC Address registers
	    MAC_SA0 = (emacETHADDR0 << 8) | emacETHADDR1;
	    MAC_SA1 = (emacETHADDR2 << 8) | emacETHADDR3;
	    MAC_SA2 = (emacETHADDR4 << 8) | emacETHADDR5;

		// Initialize Tx and Rx DMA Descriptors
		rx_descr_init ();
		tx_descr_init ();

		// Receive Broadcast and Perfect Match Packets
		MAC_RXFILTERCTRL = RFC_BCAST_EN | RFC_PERFECT_EN;

		portENTER_CRITICAL();
		{
			// Enable receive and transmit mode of MAC Ethernet core
			MAC_COMMAND  |= (CR_RX_EN | CR_TX_EN);
			MAC_MAC1     |= MAC1_REC_EN;


			prvSetupEMACInterrupt();
			vPassEMACSemaphore( xSemaphore );
		}
		portEXIT_CRITICAL();
	}

	return xSemaphore;
}
/*-----------------------------------------------------------*/
// Keil: function added to initialize Rx Descriptors
void rx_descr_init (void)
{
  unsigned int i;

  for (i = 0; i < NUM_RX_FRAG; i++) {
    RX_DESC_PACKET(i)  = RX_BUF(i);
    RX_DESC_CTRL(i)    = RCTRL_INT | (ETH_FRAG_SIZE-1);
    RX_STAT_INFO(i)    = 0;
    RX_STAT_HASHCRC(i) = 0;
  }

  /* Set EMAC Receive Descriptor Registers. */
  MAC_RXDESCRIPTOR    = RX_DESC_BASE;
  MAC_RXSTATUS        = RX_STAT_BASE;
  MAC_RXDESCRIPTORNUM = NUM_RX_FRAG-1;

  /* Rx Descriptors Point to 0 */
  MAC_RXCONSUMEINDEX  = 0;
}


// Keil: function added to initialize Tx Descriptors
void tx_descr_init (void) {
  unsigned int i;

  for (i = 0; i < NUM_TX_FRAG; i++) {
    TX_DESC_PACKET(i) = TX_BUF(i);
    TX_DESC_CTRL(i)   = 0;
    TX_STAT_INFO(i)   = 0;
  }

  /* Set EMAC Transmit Descriptor Registers. */
  MAC_TXDESCRIPTOR    = TX_DESC_BASE;
  MAC_TXSTATUS        = TX_STAT_BASE;
  MAC_TXDESCRIPTORNUM = NUM_TX_FRAG-1;

  /* Tx Descriptors Point to 0 */
  MAC_TXPRODUCEINDEX  = 0;
}
/*-----------------------------------------------------------*/

static void prvSetupEMACInterrupt( void )
{
	/* Create the semaphore used to trigger the EMAC task. */
	vSemaphoreCreateBinary( xSemaphore );
	if( xSemaphore )
	{
		/* We start by 'taking' the semaphore so the ISR can 'give' it when the
		first interrupt occurs. */
		xSemaphoreTake( xSemaphore, emacNO_DELAY );
		portENTER_CRITICAL();
		{
			// Enable EMAC interrupts.
			MAC_INTENABLE = INT_RX_DONE;

			// Reset all interrupts
			MAC_INTCLEAR  = 0xFFFF;

			// TODO: necessary to change "EMAC Priority"
			// Configure VIC for EMAC interrupt.
			VICVectAddr21 = ( portLONG ) vEMACISR_Wrapper;
			VICVectCntl21 = 5;
			VICIntEnable |= 0x00200000;
		}
		portEXIT_CRITICAL();
	}
}





/*-----------------------------------------------------------*/
// Keil: function added to write PHY
void write_PHY (int PhyReg, int Value)
{
  unsigned int tout;

  MAC_MADR = KSZ8721_DEF_ADR | PhyReg;
  MAC_MWTD = Value;

  /* Wait utill operation completed */
  tout = 0;
  for (tout = 0; tout < MII_WR_TOUT; tout++) {
    if ((MAC_MIND & MIND_BUSY) == 0) {
      break;
    }
  }
}


// Keil: function added to read PHY
unsigned short read_PHY (unsigned char PhyReg)
{
  unsigned int tout;

  MAC_MADR = KSZ8721_DEF_ADR | PhyReg;
  MAC_MCMD = MCMD_READ;

  /* Wait until operation completed */
  tout = 0;
  for (tout = 0; tout < MII_RD_TOUT; tout++) {
    if ((MAC_MIND & MIND_BUSY) == 0) {
      break;
    }
  }
  MAC_MCMD = 0;
  return (MAC_MRDD);
}
/*-----------------------------------------------------------*/

void vEMACWaitForInput( void )
{
	/* Just wait until we are signled from an ISR that data is available, or
	we simply time out. */
	xSemaphoreTake( xSemaphore, emacBLOCK_TIME_WAITING_FOR_INPUT );
}
